-- returns true if enough has time has passed or false otherwise
-- needs an id to track different calls
local next_execution_frame = {}
local function wait_frames(id, wait_frames)
  next_execution_frame[id] = next_execution_frame[id] or 0
  if next_execution_frame[id] < GameGetFrameNum() then
    next_execution_frame[id] = GameGetFrameNum() + wait_frames
    return true
  end
  return false
end

local function reset_wait_frames(id)
  next_execution_frame[id] = 0
end

local function split_string(inputstr, sep)
  sep = sep or "%s"
  local t= {}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
    table.insert(t, str)
  end
  return t
end

local function shuffle_table(t)
  local n = #t
  for i = n, 2, -1 do
    local j = Random(i)
    t[i], t[j] = t[j], t[i]
  end
end

local picked_elements = {}
local function get_random_element_sequence(t)
  if not picked_elements[t] or #picked_elements[t] == 0 then
    -- Regenerate sentences
    picked_elements[t] = {}
    for i, v in ipairs(t) do
      table.insert(picked_elements[t], v)
    end
    shuffle_table(picked_elements[t])
  end
  return table.remove(picked_elements[t], #picked_elements[t])
end

local function seed_with_frame_and_cam_pos(entity_id)
  local cam_x, cam_y = GameGetCameraPos()
  SetRandomSeed(cam_x + GameGetFrameNum(), cam_y + (entity_id or 0))
end

local function get_variable_storage_component(entity_id, name)
  for i, comp in ipairs(EntityGetComponentIncludingDisabled(entity_id, "VariableStorageComponent") or {}) do
    if ComponentGetValue2(comp, "name") == name then
      return comp
    end
  end
end

local function get_child_by_name(entity_id, name)
  for i, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
    if EntityGetName(child) == name then
      return child
    end
  end
end

-- Thanks to kaliuresis

local function lerp(a, b, weight)
	return a * weight + b * (1 - weight)
end

local function get_magnitude( x, y )
	local result = math.sqrt( x ^ 2 + y ^ 2 )
	return result
end

local function vec_normalize(x, y)
	local m = get_magnitude(x, y)
	if m == 0 then return 0,0 end
	x = x / m
	y = y / m
	return x,y
end

local function fade(t)
  --return t * t * (3 - 2*t) --cubic
  return t * t * t * (t * (t * 6 - 15) + 10) --quintic
end

-- for perlin noise
local function grad(X,Y, x, y)
    SetRandomSeed(X,Y)
    local angle = Randomf(0, 2*math.pi)
    local gx, gy = math.cos(angle), math.sin(angle)
    return gx*x+gy*y
end

-- -- Value noise
-- local function grad(X,Y, x, y)
--   SetRandomSeed(X,Y)
--   return Randomf(0,1)
-- end

local function perlin_noise(x, y)
  local X = math.floor(x)
  local Y = math.floor(y)
  x = x - math.floor(x)
  y = y - math.floor(y)
  local u = fade(x)
  local v = fade(y)
  return lerp(lerp(grad(X+1,Y+1, x - 1, y - 1), grad(X,Y+1, x, y - 1), u),
              lerp(grad(X+1,Y, x - 1, y), grad(X,Y, x, y), u),
              v)
end

-- modifies t1 in place and overwrites/adds keys/values of t1 with t2, t3, t4 etc in order
local function merge_tables_pairs(t1, ...)
  local tables = {...}
  for i, t in ipairs(tables) do
    for k, v in pairs(t) do
      t1[k] = v
    end
  end
  return t1
end

local function spawn_teleport_portal(x, y, target_x, target_y, ui_name)
  local portal = EntityLoad("mods/banana_of_rectitude/files/teleport_portal.xml", x, y)
  local teleport_comp = EntityGetFirstComponentIncludingDisabled(portal, "TeleportComponent")
  if teleport_comp then
    ComponentSetValue2(teleport_comp, "target", target_x, target_y)
  end
  local ui_info_comp = EntityGetFirstComponentIncludingDisabled(portal, "UIInfoComponent")
  if ui_info_comp then
    ComponentSetValue2(ui_info_comp, "name", ui_name)
  end
  return portal
end

local words = {
  "Rectitude",
  "Honesty",
  "Integrity",
  "Morality",
  "Virtue",
  "Ethics",
  "Uprightness",
  "Righteousness",
  "Goodness",
  "Principle",
  "Decency",
  "Fairness",
  "Justness",
  "Probity",
  "Conscience",
  "Rightness",
  "Trustworthiness",
  "Honor",
  "Justice",
  "Nobility",
  "Upstanding",
  "Conduct",
  "Purity",
  "Sincerity",
  "Dignity",
  "Respect",
  "Godliness",
  "Saintliness",
  "Fidelity",
}

local function shuffle_table_with_chance(t, chance)
  local n = #t
  for i = n, 2, -1 do
    local j = math.random(i)
    if math.random() < chance then
      t[i], t[j] = t[j], t[i]
    end
  end
end

-- We need to create a seed before the world is initialized, so getting the world seed is not possible.
-- os.time() is also not available to safe mods, so my idea was to strore the seed for the NEXT run during
-- the previous run using ModSettingSet, when the world seed IS available, and then in the next run get
-- that seed back, because ModSettingGet IS available before world init, use GameGetDateAndTimeUTC for
-- the first run
local function create_seed()
  local _,month,day,hour,minute,second = GameGetDateAndTimeUTC()
  local seed = (month * 100000 + day * 10000 + hour * 1000 + minute * 100 + second) * 1000
  return tonumber(ModSettingGet("banana_of_rectitude.random_seed")) or seed
end

-- Shuffle sprite chunks around
local function corrupt_sprite(filepath, chunk_size, corruption_amount)
  if not ModImageDoesExist(filepath) then
    error("File does not exist :" ..  tostring(filepath), 2)
  end
  local id, w, h = ModImageMakeEditable(filepath, 0, 0)
  local chunks = {}
  for y=0, h-1, chunk_size do
    for x=0, w-1, chunk_size do
      local chunk = {}
      for yy=0, chunk_size-1 do
        for xx=0, chunk_size-1 do
          if x + xx >= w or y + yy >= h then
            table.insert(chunk, 0)
          else
            local color = ModImageGetPixel(id, x + xx, y + yy)
            table.insert(chunk, color)
          end
        end
      end
      table.insert(chunks, chunk)
    end
  end
  math.randomseed(create_seed())
  shuffle_table_with_chance(chunks, corruption_amount)
  for y=0, h-1, chunk_size do
    for x=0, w-1, chunk_size do
      local chunk = chunks[math.floor(y/chunk_size) * math.ceil(w / chunk_size) + math.floor(x/chunk_size) + 1]
      for yy=0, chunk_size-1 do
        for xx=0, chunk_size-1 do
          if x + xx < w and y + yy < h then
            ModImageSetPixel(id, x + xx, y + yy, chunk[yy * chunk_size + xx + 1])
          end
        end
      end
    end
  end
end

-- Overwrite sprite chunks with random memory samples from noita
local function corrupt_sprite2(filepath, chunk_size, corruption_amount)
  if not ModImageDoesExist(filepath) then
    print("File does not exist :" ..  tostring(filepath))
    return
  end
  local id_src, src_w, src_h = ModImageMakeEditable("mods/banana_of_rectitude/files/corruption.png", 0, 0)
  local id, w, h = ModImageMakeEditable(filepath, 0, 0)
  local chunks = {}
  for y=0, src_h-1, chunk_size do
    for x=0, src_w-1, chunk_size do
      local chunk = {}
      for yy=0, chunk_size-1 do
        for xx=0, chunk_size-1 do
          if x + xx >= src_w or y + yy >= src_h then
            table.insert(chunk, 0)
          else
            local color = ModImageGetPixel(id_src, x + xx, y + yy)
            table.insert(chunk, color)
          end
        end
      end
      table.insert(chunks, chunk)
    end
  end
  shuffle_table_with_chance(chunks, 1)
  for y=0, h-1, chunk_size do
    for x=0, w-1, chunk_size do
      if math.random() < corruption_amount then
        local chunk = chunks[(math.floor(y/chunk_size) * math.ceil(src_w / chunk_size) + math.floor(x/chunk_size)) % #chunks + 1]
        for yy=0, chunk_size-1 do
          for xx=0, chunk_size-1 do
            ModImageSetPixel(id, (x + xx) % w, (y + yy) % h, chunk[yy * chunk_size + xx + 1])
          end
        end
      end
    end
  end
end

-- Add some white noise and shift some rows
local function corrupt_sprite3(filepath, chunk_size, corruption_amount)
  if not ModImageDoesExist(filepath) then
    print("File does not exist :" ..  tostring(filepath))
    return
  end
  local id, w, h = ModImageMakeEditable(filepath, 0, 0)
  local seed = create_seed()
  for i=1, 2000 do
    math.randomseed(seed + i)
    local rnd_src_offset_x = math.random(w-1)
    local rnd_src_offset_y = math.random(h-1)
    local target_x = math.random(w)
    local target_y = math.random(h)
    local noise = math.random() < 0.5
    for src_y=0, chunk_size do
      local column_offset = math.random(-3, 3)
      for src_x=0, chunk_size do
        -- ABGR
        if noise then
          local target_x = (target_x + src_x) % w
          local target_y = (target_y + src_y) % h
          local lightness = math.random(0xFF)
          ModImageSetPixel(id, target_x, target_y, math.random(0xFF) * 0x01000000 + lightness * 0x010000 + lightness * 0x0100 + lightness)
          -- ModImageSetPixel(id, target_x, target_y, 0xFF000000 + math.random(0xFF) * 0x010000 + math.random(0xFF) * 0x0100 + math.random(0xFF))
        else
          local src_x = (rnd_src_offset_x + src_x + column_offset) % w
          local src_y = (rnd_src_offset_y + src_y) % h
          local target_x = (target_x + src_x) % w
          local target_y = (target_y + src_y) % h
          local color = ModImageGetPixel(id, src_x, src_y)
          ModImageSetPixel(id, target_x, target_y, color)
        end
      end
    end
  end
end

local function clamp(value, min, max)
	value = math.max(value, min)
	value = math.min(value, max)
	return value
end

-- Adjusts a multiplier so the bonus is always beneficial
-- for instance, a 20% recharge time reduction would always go towards 0,
-- from 60 frames to 48
-- from -60 frames to -48, but this is actually detrimental, that's why in this case we need to
-- INCREASE the value from -60 to -72
-- Input: value=positive, multiplier = 0.02, returns 1.02
-- Input: value=negative, multiplier = 0.02, returns 0.98
local function adjust_multiplier(value, multiplier)
  if value < 0 then
    return 1 - multiplier
  else
    return 1 + multiplier
  end
end

-- Take a string with the test$[0-9].png syntax and returns a table with test0.png, test1.png etc
local function expand_string(str)
  local prefix, range, suffix = str:match("(.-)%$%[([%d%-]+)%](.*)")
  if not prefix or not range or not suffix then
      return {str}
  end

  local result = {}
  local start, stop = range:match("([^%-]+)%-([^%-]+)")
  if not start or not stop then
      return {str}
  end

  start, stop = tonumber(start), tonumber(stop)
  for i = start, stop do
      table.insert(result, string.format("%s%01d%s", prefix, i, suffix))
  end

  return result
end

local function world_to_screen_coords(gui, x, y)
  local ww, wh = MagicNumbersGetValue("VIRTUAL_RESOLUTION_X"), MagicNumbersGetValue("VIRTUAL_RESOLUTION_Y")
  local sw, sh = GuiGetScreenDimensions(gui)
  local _, _, cam_w, cam_h = GameGetCameraBounds()
  local cx, cy = GameGetCameraPos()
  cx = cx - cam_w / 2
  cy = cy - cam_h / 2
  x, y = x - cx, y - cy
  x, y = x / ww, y / wh
  x, y = x * sw, y * sh
  return x, y
end

local function map(value, old_min, old_max, new_min, new_max)
	return (value - old_min) * (new_max - new_min) / (old_max - old_min) + new_min;
end

local function are_floats_equal(f1, f2)
  return math.abs(f1 - f2) < 0.0001
end

local function is_point_in_rect(x, y, rx, ry, rw, rh)
  return x >= rx and x <= rx + rw and y >= ry and y <= ry + rh
end

local function get_mouse_gui_pos(gui)
  -- These seem to always be 1280, 720 no matter the actual screen size/resolution
  local mouse_screen_x, mouse_screen_y = InputGetMousePosOnScreen() -- or ControlsComponent:mMousePositionRaw
  local mx_p, my_p = mouse_screen_x / 1280, mouse_screen_y / 720
  local gui_width, gui_height = GuiGetScreenDimensions(gui)
  return mx_p * gui_width, my_p * gui_height
end

-- Gets the player whether they're polymorphed or not
local function get_player()
  local player = EntityGetWithTag("player_unit")[1]
  if not player then
    player = EntityGetWithTag("polymorphed_player")[1]
  end
  return player
end

return {
  shuffle_table = shuffle_table,
  get_random_element_sequence = get_random_element_sequence,
  wait_frames = wait_frames,
  reset_wait_frames = reset_wait_frames,
  seed_with_frame_and_cam_pos = seed_with_frame_and_cam_pos,
  get_variable_storage_component = get_variable_storage_component,
  perlin_noise = perlin_noise,
  get_child_by_name = get_child_by_name,
  merge_tables_pairs = merge_tables_pairs,
  spawn_teleport_portal = spawn_teleport_portal,
  words = words,
  corrupt_sprite = corrupt_sprite,
  corrupt_sprite2 = corrupt_sprite2,
  corrupt_sprite3 = corrupt_sprite3,
  clamp = clamp,
  create_seed = create_seed,
  adjust_multiplier = adjust_multiplier,
  expand_string = expand_string,
  world_to_screen_coords = world_to_screen_coords,
  map = map,
  are_floats_equal = are_floats_equal,
  is_point_in_rect = is_point_in_rect,
  get_mouse_gui_pos = get_mouse_gui_pos,
  split_string = split_string,
  get_player = get_player,
}
